<?php

namespace App\Classes\Parser;

use Illuminate\Support\Facades\Log;

use App\Classes\FrameFields;
use App\Classes\Parser as AbstractParser;
use App\Classes\Frame\Videotex as VideotexFrame;

class Videotex extends AbstractParser
{
	protected function parse($startline): string
	{
		// Our starting coordinates
		$output = '';

		$infield = FALSE;           // In a field
		$fieldtype = NULL;          // Type of field
		$fieldlength = 0;           // Length of field

		// $fieldadrline = 1;

		// Scan the frame for a field start
		for ($y=$startline-1;$y<=VideotexFrame::$frame_length;$y++)
		{
			// Fields can only be on a single line
			$fieldx = $fieldy = FALSE;

			for ($x=0;$x<VideotexFrame::$frame_width;$x++)
			{
				$posn = $y*VideotexFrame::$frame_width+$x;

				// If the frame is not big enough, fill it with spaces.
				$byte = ord(isset($this->content{$posn}) ? $this->content{$posn} : ' ')%128;

				// Check for start-of-field
				if ($byte == ord(ESC)) {     // Esc designates start of field (Esc-K is end of edit)
					$infield = TRUE;
					$fieldlength = 1;
					$fieldtype = ord(substr($this->content,$posn+1,1))%128;
					$output .= VideotexFrame::$if_filler;

				} else {
					if ($infield) {
						if ($byte == $fieldtype) {
							$fieldlength++;
							$byte = ord(VideotexFrame::$if_filler);	    // Replace field with VideotexFrame::$if_filler.

							if ($fieldx === FALSE) {
								$fieldx = $x;
								$fieldy = $y;
							}

							// Is this a magic field?
							if (array_get($this->fieldmap,chr($fieldtype)) ) {
								$field = $this->fieldmap[chr($fieldtype)];
								//dump(['infield','byte'=>$byte,'fieldtype'=>$fieldtype,'field'=>$field,'strpos'=>strpos($field,'#')]);

								/*
								// address field has many lines. increment when hit on first character.
								if ($fieldlength == 1 && strpos($field,'#') !== false) {
									$field = str_replace('#',$fieldadrline,$field);
									dump(['field'=>$field,'fieldadrline'=>$fieldadrline,'fieldadrline'=>$fieldadrline]);
									$fieldadrline++;
								}
								*/

								// Replace field with Date
								if ($field == '%date') {
									// Drop the last dot and replace it.
									if ($fieldlength == 2) {
										$datetime = date('D d M H:ia');
										$output = rtrim($output,VideotexFrame::$if_filler);
										$output .= $datetime{0};
									}

									if ($fieldlength > 1 AND $fieldlength <= strlen($datetime))
										$byte = ord($datetime{$fieldlength-1});
								}

								// @todo user data
								/* else if (isset($user[$field])) {
									if ($fieldlength <= strlen($user[$field])) {
										$byte = ord($user[$field]{$fieldlength-1});
									}
								} /*else 	// pre-load field contents. PAM or *00 ?
								if (isset($fields[what]['value'])) {

								*/
							}

						} else {
							$this->fields->push(new FrameFields([
								'type'=>chr($fieldtype),
								'length'=>$fieldlength,
								'x'=>$fieldx-1,             // Adjust for the ESC char
								'y'=>$fieldy,
							]));

							Log::debug(sprintf('Frame Field found at [%s,%s], Type: %s, Length: %s',$fieldx-1,$fieldy,$fieldtype,$fieldlength));

							$infield = FALSE;
							$fieldx = $fieldy = FALSE;
						}
					}
				}

				// truncate end of lines @todo havent validated this code or used it?
				if (isset($pageflags['tru']) && substr($this->content,$posn,40-$x) === str_repeat(' ',40-$x)) {
					$output .= CR . LF;
					break;
				}

				if (! $infield OR $fieldlength > 1)
					$output .= ($byte < 32) ? ESC.chr($byte+64) : chr($byte);
			}
		}

		return $output;
	}
}